@@ -1,11 +1,16 @@
 """
 unmatcher :: Regular expression reverser for Python
 """
-__version__ = "0.0.2"
+__version__ = "0.1"
 __author__ = "Karol Kuczmarski"
 __license__ = "Simplified BSD"
 
 
+try:
+    from itertools import imap
+except ImportError:
+    imap = map  # Python 3
+
 import random
 import re
 import string
@@ -83,7 +88,7 @@ class Reversal(object):
 
     # TODO: choose among Unicode characters if using Unicode
     BUILTIN_CHARSETS = {
-        'word': string.ascii_letters,
+        'word': string.ascii_letters + string.digits,
         'digit': string.digits,
         'space': string.whitespace,
     }
@@ -113,17 +118,19 @@ class Reversal(object):
         """Generates string matching given sequence of nodes
         from regular expressions' abstract syntax tree (AST).
         """
-        return self._str().join(map(self._reverse_node, nodes))
+        return self._str().join(imap(self._reverse_node, nodes))
 
     def _reverse_node(self, (type_, data)):
         """Generates string matching given node from regular expression AST."""
         if type_ == 'literal':
             return self._chr(data)
+        if type_ == 'not_literal':  # [^X], where X ia a character
+            return random.choice(self._negate(self._chr(data)))
         if type_ == 'any':
             return random.choice(self._charset('any'))
 
         if type_ == 'in':
-            return self._reverse_choice_node(data)
+            return self._reverse_charset_node(data)
         if type_ == 'branch':
             return self._reverse_branch_node(data)
 
@@ -143,51 +150,37 @@ class Reversal(object):
         raise NotImplementedError(
             "unsupported regular expression element: %s" % type_)
 
-    def _reverse_choice_node(self, node_data):
+    def _reverse_charset_node(self, node_data):
         """Generates string matching 'in' node from regular expr. AST.
 
-        This node is an alternative between several variants (child nodes),
-        incl. some that are not encountered in other places
-        (character sets, for example).
+        This node matches a specified set of characters. Typically,
+        it is expressed using the ``[...]`` notation, but it can also arise
+        from simple uses of ``|`` operator, where all branches match
+        just one, literal character (e.g. ``a|b|c``).
         """
-        first, _ = node_data[0]
-        if first == 'negate':
-            # TODO: support this
-            raise NotImplementedError(
-                "negated character sets [^...] are not supported")
-
-        # TODO: charset variants might be of different size
-        # (wrt to no. of chars they match), but they are all assigned
-        # the same weight in the random.choice() below (along with non-charset
-        # variants) - e.g. [a-cd] has the same chance to pick `a-c` or `d`;
-        # see about correcting that (how about non-charset variants, though?)
-        chosen = random.choice(node_data)
-        type_, data = chosen
-
-        # range (e.g. a-z) inside [ ]
-        if type_ == 'range':
-            min_char, max_char = data
-            return self._chr(random.randint(min_char, max_char))
-
-        # built-in character set: \d, \w, etc.
-        if type_ == 'category':
-            return self._reverse_builtin_charset_node(data)
-
-        return self._reverse_node(chosen)
-
-    def _reverse_builtin_charset_node(self, node_data):
-        """Generates string matching 'category' node from regular expr. AST.
-
-        This node matches a built-in set of characters, like ``\d`` or ``\w``.
-        """
-        _, type_ = node_data.rsplit('_', 1)  # category(_not)?_(digit|word|etc)
-        negate = '_not_' in node_data
-
-        charset = self._charset(type_)
+        negate = node_data[0][0] == 'negate'
         if negate:
-            all_chars = self._charset('any', flags=0)
-            charset = list(set(all_chars) - set(charset))
-        return random.choice(charset)
+            node_data = node_data[1:]
+
+        charset = set()
+        for type_, data in node_data:
+            if type_ == 'literal':
+                charset.add(self._chr(data))
+            elif type_ == 'range':
+                min_char, max_char = data
+                charset.update(imap(self._chr, xrange(min_char, max_char + 1)))
+            elif type_ == 'category':
+                _, what = data.rsplit('_', 1)  # category(_not)?_(digit|word|etc)
+                category_chars = self._charset(what)
+                if '_not_' in data:
+                    category_chars = self._negate(category_chars)
+                charset.update(category_chars)
+            else:
+                raise ValueError("invalid charset alternative: %s" % type_)
+
+        if negate:
+            charset = self._negate(charset)
+        return random.choice(list(charset))
 
     def _reverse_repeat_node(self, node_data):
         """Generates string matching 'min_repeat' or 'max_repeat' node
@@ -246,7 +239,7 @@ class Reversal(object):
         index = node_data
         return self.groups[index]
 
-    # Utility functions
+    # Handling character sets
 
     def _charset(self, name, flags=None):
         """Return chars belonging to charset of given name.
@@ -267,3 +260,8 @@ class Reversal(object):
             return self.BUILTIN_CHARSETS[name]
 
         raise ValueError("invalid charset name '%s'" % name)
+
+    def _negate(self, charset):
+        """Returns negated version of given charset."""
+        all_chars = self._charset('any', flags=0)
+        return list(set(all_chars) - set(charset))
